/*
 * Copyright (c) 1996-2009 Barton P. Miller
 * 
 * We provide the Paradyn Parallel Performance Tools (below
 * described as "Paradyn") on an AS IS basis, and do not warrant its
 * validity or performance.  We reserve the right to update, modify,
 * or discontinue this software at any time.  We shall have no
 * obligation to supply such updates or modifications or any other
 * form of support to you.
 * 
 * By your use of Paradyn, you understand and agree that we (or any
 * other person or entity with proprietary rights in Paradyn) are
 * under no obligation to provide either maintenance services,
 * update services, notices of latent defects, or correction of
 * defects for Paradyn.
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 */

// $Id: BPatch_image.C,v 1.120 2008/06/20 22:00:02 legendre Exp $

#define BPATCH_FILE

#include <stdio.h>
#include <assert.h>
#include <string.h>
#include <string>

#include "process.h"
#include "instPoint.h"
#include "instP.h"
#include "function.h"

#include "mapped_module.h"
#include "mapped_object.h"

#include "BPatch.h"
#include "BPatch_image.h"
#include "BPatch_type.h"
#include "BPatch_collections.h"
#include "BPatch_libInfo.h"
#include "BPatch_statement.h"
#include "BPatch_function.h" 

#include "addressSpace.h"

#include "ast.h"
using namespace Dyninst;
using namespace std;

/*
 * BPatch_image::BPatch_image
 *
 * Construct a BPatch_image for the given process.
 */

BPatch_image::BPatch_image(BPatch_addressSpace *_addSpace) :
   addSpace(_addSpace)
{
   _srcType = BPatch_sourceProgram;
}

/*
 * BPatch_image::BPatch_image
 *
 * Construct a BPatch_image.
 */
BPatch_image::BPatch_image() :
   addSpace(NULL)
{
   _srcType = BPatch_sourceProgram;
}

/* 
 * Cleanup the image's memory usage when done.
 *
 */
BPatch_image::~BPatch_image()
{
   for (unsigned int i = 0; i < modlist.size(); i++) {
      delete modlist[i];
   }

   for (unsigned j = 0; j < removed_list.size(); j++) {
      delete removed_list[j];
   }
}


/*
 * getThr - Return the BPatch_thread
 *
 */
BPatch_thread *BPatch_image::getThrInt()
{
   assert(addSpace->getType() == TRADITIONAL_PROCESS);
   BPatch_process * bpTemp = dynamic_cast<BPatch_process *>(addSpace);

   assert(bpTemp->threads.size() > 0);
   return bpTemp->threads[0];
}

BPatch_process *BPatch_image::getProcessInt()
{
   if (addSpace->getType() == TRADITIONAL_PROCESS)
      return dynamic_cast<BPatch_process *>(addSpace);
   else
      return NULL;   
}

BPatch_addressSpace *BPatch_image::getAddressSpaceInt()
{
   return addSpace;
}

/* 
 * getSourceObj - Return the children (modules)
 *
 */
bool BPatch_image::getSourceObjInt(BPatch_Vector<BPatch_sourceObj *> &vect)
{
   BPatch_Vector<BPatch_module *> *temp = getModules();
   if (temp) {
      vect = * (BPatch_Vector<BPatch_sourceObj *> *) temp;
      return (true);
   } else {
      return (false);
   }
}

/* 
 * getObjParent - Return the parent (this is the top level so its null)
 *
 */

BPatch_sourceObj *BPatch_image::getObjParentInt()
{
   return NULL;
}

/*
 * BPatch_image::getProcedures
 *
 * Returns a list of all procedures in the image upon success, and NULL
 * upon failure.
 */

BPatch_Vector<BPatch_function *> *BPatch_image::getProceduresInt(bool incUninstrumentable)
{
   BPatch_Vector<BPatch_function *> *proclist = new BPatch_Vector<BPatch_function *>;
   bool some_succeeded = false;
   if (proclist == NULL) return NULL;

   BPatch_Vector<BPatch_module *> *mods = getModules();

   for (unsigned int i = 0; i < (unsigned) mods->size(); i++) {
      bool result = (*mods)[i]->getProcedures(*proclist, incUninstrumentable);
      if (result)
         some_succeeded = true;
   }
   if (!some_succeeded) {
      delete proclist;
      return NULL;
   }
   
   return proclist;
}

bool BPatch_image::getProceduresInt(BPatch_Vector<BPatch_function*> &procs, bool incUninstrumentable)
{
   bool some_succeeded = false;
   BPatch_Vector<BPatch_module *> *mods = getModules();

   for (unsigned int i = 0; i < (unsigned) mods->size(); i++) {
      bool result = (*mods)[i]->getProcedures(procs, incUninstrumentable);
      if (result)
         some_succeeded = true;
   }
   return some_succeeded;
}


BPatch_Vector<BPatch_parRegion *> *BPatch_image::getParRegionsInt(bool incUninstrumentable)
{
   BPatch_Vector<BPatch_function *> procList;
   bool result = getProcedures(procList, incUninstrumentable);
   if (!result)
      return NULL;

   BPatch_Vector<BPatch_parRegion *> *parRegionList = new BPatch_Vector<BPatch_parRegion *>;
   if (parRegionList == NULL) return NULL;


   for (unsigned int i=0; i < (unsigned) procList.size(); i++) {
      int_function * intF = procList[i]->lowlevel_func();
      const pdvector<int_parRegion *> intPR = intF->parRegions();

      for (unsigned int j=0; j < (unsigned) intPR.size(); j++)
      {
         BPatch_parRegion * pR = new BPatch_parRegion(intPR[j], procList[i]);
         parRegionList->push_back(pR);
      }
   }

   return parRegionList;
}

/*
 * BPatch_image::getGlobalVariables
 *
 * Returns a list of all variables in the image upon success, and NULL
 * upon failure.
 */
BPatch_Vector<BPatch_variableExpr *> *BPatch_image::getGlobalVariablesInt()
{
   if (!addSpace)
      return NULL;

   BPatch_Vector<BPatch_variableExpr *> *varlist = new BPatch_Vector<BPatch_variableExpr *>;
   if (!varlist) return NULL;

   bool result = getVariablesInt(*varlist);
   if (!result) {
      delete varlist;
      return NULL;
   }
   return varlist;
}

bool BPatch_image::getVariablesInt(BPatch_Vector<BPatch_variableExpr *> &vars)
{
   bool some_succeeded = false;

   if (!addSpace)
      return false;

   BPatch_Vector<BPatch_module *> mods;
   getModules(mods);
   
   for (unsigned int m = 0; m < mods.size(); m++) {
      BPatch_module *module = mods[m];
      
      bool result = module->getVariables(vars);
      if (result)
         some_succeeded = true;
   }

   return some_succeeded;
}

bool BPatch_image::setFuncModulesCallback(BPatch_function *bpf, void *data)
{
   BPatch_image *img = (BPatch_image *) data;
   if (bpf->getModule() == NULL && bpf->func->mod() != NULL) {
      bpf->mod = img->findModule(bpf->func->mod()->fileName().c_str());
   }

   if ( bpf->getModule() == NULL ) {
      char name[256];
      fprintf(stderr, "Warning: bpf '%s' unclaimed, setting to DEFAULT_MODULE\n",
            bpf->getName( name, 255 ) );
      bpf->setModule( img->defaultModule );
   }   
   return true;
}

bool BPatch_image::getModulesInt(BPatch_Vector<BPatch_module *> &mods)
{
   BPatch_Vector<BPatch_module *> *ret = getModulesInt();
   if (!ret) {
      return false;
   }
   mods = *ret;
   return true;
}

BPatch_Vector<BPatch_module *> *BPatch_image::getModulesInt() 
{
   pdvector<mapped_module *> modules;
   std::vector<AddressSpace *> as;
   addSpace->getAS(as);
   
   for (unsigned i=0; i<as.size(); i++) {
      as[i]->getAllModules(modules);

      // We may have created a singleton module already -- check to see that we 
      // don't double-create
      for (unsigned i = 0; i < modules.size(); i++ ) {
         mapped_module *map_mod = modules[i];
         findOrCreateModule(map_mod);
      }
   }

   return &modlist;
} 

/*
 * BPatch_image::findModule
 *
 * Returns module with <name>, NULL if not found
 */
BPatch_module *BPatch_image::findModuleInt(const char *name, bool substring_match) 
{
   char buf[512];

   if (!name) {
      bperr("%s[%d]:  findModule:  no module name provided\n",
            __FILE__, __LINE__);
      return NULL;
   }

   BPatch_module *target = NULL;

   for (unsigned int i = 0; i < modlist.size(); ++i) {
      BPatch_module *mod = modlist[i];
      assert(mod);
      mod->getName(buf, 512); 
      if (substring_match) { 
         if (strstr(buf, name)) {
            target = mod;
            break;
         }
      }
      else  {//exact match required
         if (!strcmp(name, buf)) {
            target = mod;
            break;
         }
      }
   }

   if (target) 
      return target;

   // process::findModule does a wildcard match, not a substring match 

   char *tmp = (char *) malloc(strlen(name) + 3);
   if (substring_match)
      sprintf(tmp, "*%s*", name); 
   else 
      sprintf(tmp, "%s", name);

   std::vector<AddressSpace *> as;
   addSpace->getAS(as);
   
   mapped_module *mod = NULL;
   for (unsigned i=0; i<as.size(); i++) {
      mod = as[i]->findModule(tmp,substring_match);
      if (mod)
         break;
   }

   free(tmp);
   if (!mod) return false;

   target = findOrCreateModule(mod);
   return target;
}

/*
 * BPatch_image::createInstPointAtAddr
 *
 * Returns a pointer to a BPatch_point object representing an
 * instrumentation point at the given address.
 *
 * Returns the pointer to the BPatch_point on success, or NULL upon
 * failure.
 *
 * address	The address that the instrumenation point should refer to.
 */

BPatch_point *BPatch_image::createInstPointAtAddrInt(void *address)
{
   return createInstPointAtAddr(address,NULL);
}

/*
 * BPatch_image::createInstPointAtAddr
 *
 * Returns a pointer to a BPatch_point object representing an
 * instrumentation point at the given address. If the BPatch_function
 * argument is given it has to be the function that address belongs to or NULL.
 * The function is used to bypass the function that the address belongs to
 * The alternative argument is used to retrieve the point if the new point
 * intersects with another already existing one.
 *
 * Returns the pointer to the BPatch_point on success, or NULL upon
 * failure.
 *
 * address	The address that the instrumenation point should refer to.
 */

BPatch_point *BPatch_image::createInstPointAtAddrWithAlt(void *address,
      BPatch_point** alternative,
      BPatch_function* bpf)
{
   Address address_int = (Address) address;

   unsigned i;

   std::vector<AddressSpace *> as;
   addSpace->getAS(as);
   assert(as.size());   
   AddressSpace *llAS = as[0];

   int_function *func = NULL;

   if (bpf) {
      func = bpf->func;
   }
   else {
      func = llAS->findFuncByAddr(address_int);
   }

   if (func == NULL) return NULL;

   /* See if there is an instPoint at this address */
   instPoint *p = NULL;

   if ((p = func->findInstPByAddr(address_int))) {
      return addSpace->findOrCreateBPPoint(NULL, p, BPatch_locInstruction);
   }

   /* Look in the regular instPoints of the enclosing function. */
   /* This has an interesting side effect: "instrument the first
      instruction" may return with "entry instrumentation", which can
      have different semantics. */

   // If it's in an uninstrumentable function, just return an error.
   if (!func->isInstrumentable()) { 
      return NULL;
   }

   const pdvector<instPoint *> entries = func->funcEntries();
   for (unsigned t = 0; t < entries.size(); t++) {
      assert(entries[t]);
      if (entries[t]->match(address_int)) {
         return addSpace->findOrCreateBPPoint(NULL, entries[t], BPatch_entry);
      }
   }

   const pdvector<instPoint*> &exits = func->funcExits();
   for (i = 0; i < exits.size(); i++) {
      assert(exits[i]);
      if (exits[i]->match(address_int)) {
         return addSpace->findOrCreateBPPoint(NULL, exits[i], BPatch_exit);
      }
   }

   const pdvector<instPoint*> &calls = func->funcCalls();
   for (i = 0; i < calls.size(); i++) {
      assert(calls[i]);
      if (calls[i]->match(address_int))  {
         return addSpace->findOrCreateBPPoint(NULL, calls[i], BPatch_subroutine);
      }
   }

   if (alternative)
      *alternative = NULL;

   /* We don't have an instPoint for this address, so make one. */
   instPoint *newInstP = instPoint::createArbitraryInstPoint(address_int, llAS, func);
   if (!newInstP) return NULL;

   return addSpace->findOrCreateBPPoint(NULL, newInstP, BPatch_locInstruction);
}

/*
 * BPatch_image::findFunction
 *
 * Fills a vector with BPatch_function pointers representing all functions in
 * the image with the given name.  Returns a pointer to the vector that was
 * passed in on success, and NULL on error.
 *
 * name		The name of function to look up.
 * funcs	The vector in which to place the results.
 */

BPatch_Vector<BPatch_function*> *BPatch_image::findFunctionInt(const char *name, 
      BPatch_Vector<BPatch_function*> &funcs, 
      bool showError,
      bool regex_case_sensitive,
      bool incUninstrumentable)
{
   std::vector<AddressSpace *> as;
   addSpace->getAS(as);
   assert(as.size());

   if (NULL == strpbrk(name, REGEX_CHARSET)) {
      //  usual case, no regex
      pdvector<int_function *> foundIntFuncs;
      for (unsigned i=0; i<as.size(); i++) {
         as[i]->findFuncsByAll(std::string(name), foundIntFuncs);
      }
      if (!foundIntFuncs.size())
      {
         // Error callback...
         if (showError) {
            std::string msg = std::string("Image: Unable to find function: ") + 
               std::string(name);
            BPatch_reportError(BPatchSerious, 100, msg.c_str());
         }
         return NULL;
      }
      // We have a list; if we don't want to include uninstrumentable,
      // scan and check
      for (unsigned int fi = 0; fi < foundIntFuncs.size(); fi++) {
         if (foundIntFuncs[fi]->isInstrumentable() || incUninstrumentable) {
            BPatch_function *foo = addSpace->findOrCreateBPFunc(foundIntFuncs[fi], NULL);
            funcs.push_back(foo);
         }
      }

      if (funcs.size() > 0)
         return &funcs;
      if (showError) {
         std::string msg = std::string("Image: Unable to find function: ") + 
            std::string(name);
         BPatch_reportError(BPatchSerious, 100, msg.c_str());
      }
      return NULL;
   }

#if !defined(os_windows)
   // REGEX falls through:
   regex_t comp_pat;
   int err, cflags = REG_NOSUB | REG_EXTENDED;

   if ( !regex_case_sensitive )
      cflags |= REG_ICASE;

   //cerr << "compiling regex: " <<name<<endl;

   if (0 != (err = regcomp( &comp_pat, name, cflags ))) {
      char errbuf[80];
      regerror( err, &comp_pat, errbuf, 80 );
      if (showError) {
         cerr << __FILE__ << ":" << __LINE__ << ":  REGEXEC ERROR: "<< errbuf << endl;
         std::string msg = std::string("Image: Unable to find function pattern: ") 
            + std::string(name) + ": regex error --" + std::string(errbuf);
         BPatch_reportError(BPatchSerious, 100, msg.c_str());
      }
      // remove this line
      cerr << __FILE__ << ":" << __LINE__ << ":  REGEXEC ERROR: "<< errbuf << endl;
      return NULL;
   }

   // Regular expression search. This used to be handled at the image
   // class level, but was moved up here to simplify semantics. We
   // have to iterate over every function known to the process at some
   // point, so it might as well be top-level. This is also an
   // excellent candidate for a "value-added" library.

   pdvector<int_function *> all_funcs;
   for (unsigned i=0; i<as.size(); i++) {
      as[i]->getAllFunctions(all_funcs);
   }

   for (unsigned ai = 0; ai < all_funcs.size(); ai++) {
      int_function *func = all_funcs[ai];
      // If it matches, push onto the vector
      // Check all pretty names (and then all mangled names if 
      // there is no match)
      bool found_match = false;
      unsigned piter, miter;
      for (piter = 0; piter < func->prettyNameVector().size(); piter++) {
         const string &pName = func->prettyNameVector()[piter];
         int err;

         if (0 == (err = regexec(&comp_pat, pName.c_str(), 1, NULL, 0 ))){
            if (func->isInstrumentable() || incUninstrumentable) {
               BPatch_function *foo = addSpace->findOrCreateBPFunc(func,NULL);
               //BPatch_function *foo = proc->findOrCreateBPFunc(func, NULL);
               funcs.push_back(foo);
            }
            found_match = true;
            break;
         }
      }
      if (found_match) continue; // Don't check mangled names

      for (miter = 0; miter < func->symTabNameVector().size(); miter++) {
         const string &mName = func->symTabNameVector()[miter];
         int err;

         if (0 == (err = regexec(&comp_pat, mName.c_str(), 1, NULL, 0 ))){
            if (func->isInstrumentable() || incUninstrumentable) {
               BPatch_function *foo = addSpace->findOrCreateBPFunc(func,NULL);
               //	   BPatch_function *foo = proc->findOrCreateBPFunc(func, NULL);
               funcs.push_back(foo);
            }
            found_match = true;
            break;
         }
      }
   }

   regfree(&comp_pat);

   if (funcs.size() > 0) {
      return &funcs;
   } 

   if (showError) {
      std::string msg = std::string("Unable to find pattern: ") + std::string(name);
      BPatch_reportError(BPatchSerious, 100, msg.c_str());
   }
#endif
   return NULL;
}

#ifdef IBM_BPATCH_COMPAT
//  a wrapper for IBM compatibility.  The function vector is assumed to be
//  allocated before calling.
BPatch_Vector<BPatch_function*> *BPatch_image::findFunctionPtr(
      const char *name, BPatch_Vector<BPatch_function*> *funcs, bool showError,
      bool regex_case_sensitive, bool incUninstrumentable)
{
   return findFunction(name, *funcs, showError, regex_case_sensitive, incUninstrumentable);
}
#endif

/*
 * BPatch_image::findFunction 2
 *
 * Fills a vector with BPatch_function pointers representing all functions in
 * the image according to the user defined sieving function.  
 * Returns a pointer to the vector that was passed in on success, and NULL on error.
 * 
 *
 * bpsieve      User-provided boolean function used to determine inclusion in the
 *              filtered set.
 * user_data    a pointer to a user-defined data space for use by bpsieve
 * funcs	The vector in which to place the results.
 */

   BPatch_Vector<BPatch_function *> *
BPatch_image::findFunctionWithSieve(BPatch_Vector<BPatch_function *> &funcs, 
      BPatchFunctionNameSieve bpsieve,
      void *user_data, int showError, 
      bool incUninstrumentable)
{
   pdvector<int_function *> all_funcs;
   std::vector<AddressSpace *> as;
   addSpace->getAS(as);
   assert(as.size());

   for (unsigned i=0; i<as.size(); i++)
      as[i]->getAllFunctions(all_funcs);

   for (unsigned ai = 0; ai < all_funcs.size(); ai++) {
      int_function *func = all_funcs[ai];
      // If it matches, push onto the vector
      // Check all pretty names (and then all mangled names if there is no match)
      bool found_match = false;

      for (unsigned piter = 0; piter < func->prettyNameVector().size(); piter++) {
         const string &pName = func->prettyNameVector()[piter];

         if ((*bpsieve)(pName.c_str(), user_data)) {
            if (func->isInstrumentable() || incUninstrumentable) {
               BPatch_function *foo = addSpace->findOrCreateBPFunc(func,NULL);
               //BPatch_function *foo = proc->findOrCreateBPFunc(func, NULL);
               funcs.push_back(foo);
            }
            found_match = true;
            break;
         }
      }

      if (found_match) continue; // Don't check mangled names

#if 0
      // Apparently don't check mangled at all

      for (unsigned miter = 0; miter < func->symTabNameVector().size(); miter++) {
         const std::string &mName = func->symTabNameVector()[miter];
         int err;

         if (0 == (err = regexec(&comp_pat, mName.c_str(), 1, NULL, 0 ))){
            if (func->isInstrumentable() || incUninstrumentable) {
               BPatch_function *foo = proc->findOrCreateBPFunc(func, NULL);
               funcs.push_back(foo);
            }
            found_match = true;
            break;
         }
      }
#endif
   }

   if (funcs.size() > 0) {
      return &funcs;
   } 

   if (showError) {
      const char *msg = "No function matches for sieve provided";
      BPatch_reportError(BPatchSerious, 100, msg);
   }

   return NULL;
}

/*
 * BPatch_image::findFunctionP
 *
 * Finds a function based on an address in the mutatee
 */

BPatch_function *BPatch_image::findFunctionInt(unsigned long addr)
{
   std::vector<AddressSpace *> as;
   addSpace->getAS(as);
   assert(as.size());
   int_function *ifunc = as[0]->findFuncByAddr(addr);
   if (!ifunc)
      return NULL;

   return addSpace->findOrCreateBPFunc(ifunc,NULL);
}

bool BPatch_image::findFunctionInt(Dyninst::Address addr, 
                                   BPatch_Vector<BPatch_function *> &funcs)
{
   std::vector<AddressSpace *> as;
   std::vector<int_function *> ifuncs;
   bool result;

   addSpace->getAS(as);
   assert(as.size());

   result = as[0]->findFuncsByAddr(addr, ifuncs);
   if (!result)
      return false;

   assert(ifuncs.size());
   for (unsigned i=0; i<ifuncs.size(); i++)
   {
      BPatch_function *bpfunc = addSpace->findOrCreateBPFunc(ifuncs[i], NULL);
      funcs.push_back(bpfunc);
   }

   return true;
}


/*
 * BPatch_image::findVariable
 *
 * Returns a BPatch_variableExpr* representing the given variable in the
 * application image.  If no such variable exists, returns NULL.
 *
 * name		The name of the variable to look up.
 *
 * First look for the name with an `_' prepended to it, and if that is not
 *   found try the original name.
 */
BPatch_variableExpr *BPatch_image::findVariableInt(const char *name, 
                                                   bool showError)
{
    pdvector<int_variable *> vars;
    std::vector<AddressSpace *> as;
    int_variable *var = NULL;
    AddressSpace *var_as = NULL;
    
    if (!addSpace)
       return NULL;
    
    addSpace->getAS(as);

    for (unsigned i=0; i<as.size(); i++) {
       as[i]->findVarsByAll(name, vars);
       if (vars.size()) {
          var = vars[0];
          var_as = as[i];
          break;
       }
    }

    if (!var) {
       std::string under_name = std::string("_") + std::string(name);
       for (unsigned i=0; i<as.size(); i++) {
          as[i]->findVarsByAll(under_name, vars);
          if (vars.size()) {
             var = vars[0];
             var_as = as[i];
             break;
          }
       }
    }
       
    if (!var) {
       for (unsigned i=0; i<as.size(); i++) {
          string defaultNamespacePref = as[i]->getAOut()->parse_img()->
             getObject()->getDefaultNamespacePrefix();
          string prefix_name = defaultNamespacePref + string(".") + string(name);
          as[i]->findVarsByAll(prefix_name, vars);
          if (vars.size()) {
             var = vars[0];
             var_as = as[i];
             break;
          }
       }
    }

    if (!var) {
       if (showError) {
          string msg = std::string("Unable to find variable: ") + 
             string(name);
          showErrorCallback(100, msg);
       }
       return NULL;
    }

   BPatch_variableExpr *bpvar = addSpace->findOrCreateVariable(var);
   assert(bpvar);
   return bpvar;
}

//
// findVariable
//	scp	- a BPatch_point that defines the scope of the current search
//	name	- name of the variable to find.
//

BPatch_variableExpr *BPatch_image::findVariableInScope(BPatch_point &scp,
		const char *name)
{
	// Get the function to search for it's local variables.
	// XXX - should really use more detailed scoping info here - jkh 6/30/99
	BPatch_function *func = const_cast<BPatch_function *> (scp.getFunction());

	if (!func) 
	{
		std::string msg = std::string("point passed to findVariable lacks a function\n address point type passed?");
		showErrorCallback(100, msg);
		return NULL;
	}

	AddressSpace *as = func->lladdSpace;

	BPatch_localVar *lv = func->findLocalVar(name);

	if (!lv) 
	{
		// look for it in the parameter scope now
		lv = func->findLocalParam(name);
	}

	if (lv) 
	{
		return new BPatch_variableExpr(addSpace, as, lv, lv->getType(), &scp); 
	}

	// finally check the global scope.
	// return findVariable(name);

	/* If we have something else to try, don't report errors on this failure. */
	bool reportErrors = true;
	char mangledName[100];
	func->getName( mangledName, 100 );
	char * lastScoping = NULL;      

	if ( strrchr( mangledName, ':' ) != NULL ) 
	{ 
		reportErrors = false; 
	}

	BPatch_variableExpr * gsVar = findVariable( name, reportErrors );

	if ( gsVar == NULL ) 
	{
		/* Try finding it with the function's scope prefixed. */

		if ( (lastScoping = strrchr( mangledName, ':' )) != NULL ) 
		{
			* (lastScoping + sizeof(char)) = '\0';
			char scopedName[200];
			memmove( scopedName, mangledName, strlen( mangledName ) );
			memmove( scopedName + strlen( mangledName ), name, strlen( name ) );
			scopedName[ strlen( mangledName ) + strlen( name ) ] = '\0';
			bperr( "Searching for scoped name '%s'\n", scopedName );
			gsVar = findVariable( scopedName ); 
		}
	}
	return gsVar;
}

/*
 * BPatch_image::findType
 *
 * Returns a BPatch_type* representing the named type.  If no such type
 * exists, returns NULL.
 *
 * name		The name of type to look up.
 */

BPatch_type *BPatch_image::findTypeInt(const char *name)
{
    BPatch_type *type;

    assert(BPatch::bpatch != NULL);

    // XXX - should this stuff really be by image ??? jkh 3/19/99
    BPatch_Vector<BPatch_module *> *mods = getModules();
    for (int m = mods->size() -1; m >= 0; m--) {
	BPatch_module *module = (*mods)[m];
	type = module->getModuleTypes()->findType(name);
	if (type) {
           return type;
        }
    }

    // check the default base types
    type = BPatch::bpatch->stdTypes->findType(name);
    if (type)  {
      return type;
    }
    // check the API types of last resort
    type = BPatch::bpatch->APITypes->findType(name);

    return type;

}

bool BPatch_image::getSourceLinesInt(unsigned long addr, 
                                     BPatch_Vector<BPatch_statement> &lines )
{
   unsigned int originalSize = lines.size();
   BPatch_Vector< BPatch_module * > * modules = getModules();

   /* Iteratate over the modules, looking for addr in each. */

   for (unsigned int i = 0; i < modules->size(); i++ ) 
   {
      BPatch_module *m = (*modules)[i];
      //  address-to-offset conversion done in module routine 
      m->getSourceLinesInt(addr,lines);
   }

   if ( lines.size() != originalSize ) 
   {
      return true; 
   }	

   //fprintf(stderr, "%s[%d]:  fail to getSourceLines for addr %lu\n", FILE__, __LINE__, addr);

   return false;
} /* end getSourceLines() */

bool BPatch_image::getAddressRangesInt( const char * lineSource, 
      unsigned int lineNo, 
      std::vector< std::pair<unsigned long, unsigned long> > & ranges ) 
{
   unsigned int originalSize = ranges.size();

   //  First check to see if we can find the named module
   //  If so, use it for the lookup

   BPatch_module *target_mod = findModuleInt(lineSource, false);

   if (target_mod)
   {
      target_mod->getAddressRanges(lineSource, lineNo, ranges);
      if ( ranges.size() != originalSize ) 
      {
         return true; 
      }
   }

   //  If we cannot find the given module, try (perhaps against all hope)
   //  looking in all the modules sequentially

   BPatch_Vector< BPatch_module * > * modules = getModules();

   // Iteratate over the modules, looking for ranges in each. 

   for ( unsigned int i = 0; i < modules->size(); i++ ) 
   {
      BPatch_module *m = (*modules)[i];
      m->getAddressRanges(lineSource, lineNo, ranges);
   }

   if ( ranges.size() != originalSize ) 
   {
      return true; 
   }

   dwarf_printf("%s[%d]:  getAddressRanges failing\n", FILE__, __LINE__);
   return false;
} 


char *BPatch_image::getProgramNameInt(char *name, unsigned int len) 
{
   std::vector<AddressSpace*> as;
   addSpace->getAS(as);
   AddressSpace *aout = as[0];

   if (!aout->mappedObjects().size()) {
      strncpy(name, "<no program defined>", len);
   }

   const char *imname =  aout->getAOut()->fullName().c_str();
   if (NULL == imname) imname = "<unnamed image>";

   strncpy(name, imname, len);
   return name;
}

char *BPatch_image::getProgramFileNameInt(char *name, unsigned int len)
{
   std::vector<AddressSpace*> as;
   addSpace->getAS(as);
   AddressSpace *aout = as[0];

   if (!aout->mappedObjects().size()) {
      // No program defined yet
      strncpy(name, "<no program defined>", len);
   }

   const char *imname =  aout->getAOut()->fileName().c_str();
   if (NULL == imname) imname = "<unnamed image file>";

   strncpy(name, imname, len);
   return name;
}

#ifdef IBM_BPATCH_COMPAT
char *BPatch_image::programNameInt(char *name, unsigned int len) 
{
   return getProgramName(name, len);
}

int  BPatch_image::lpTypeInt() 
{
   switch(addSpace->getAS()->getAddressWidth()) {
      case 4: return LP32; break;
      case 8: return LP64; break;
      default: return UNKNOWN_LP; break;
   }
};
#endif

BPatch_module *BPatch_image::findModule(mapped_module *base) 
{
   // As below, but don't create if none already exists

   BPatch_module *bpm = NULL;
   for (unsigned j = 0; j < modlist.size(); j++) {
      if (modlist[j]->lowlevel_mod() == base) {
         bpm = modlist[j];
         break;
      }
   }
   return bpm;
}

BPatch_module *BPatch_image::findOrCreateModule(mapped_module *base) 
{
   BPatch_module *bpm = findModule(base);
   
   if (bpm == NULL) {
      bpm = new BPatch_module( addSpace, base->proc(), base, this );
      modlist.push_back( bpm );
   }
   assert(bpm != NULL);
   return bpm;
}


/* This function should only be invoked when regionStart and regionEnd
 * have been verified as not conflicting with existing mapped objects,
 * and when it has also been verified that all elements of
 * funcEntryAddrs lie between the region regionStart and regionEnd
 */
BPatch_module *parseRegion(BPatch_process *bpProc, 
                           const Address regionStart, const Address regionEnd, 
                           const vector<Address> &funcEntryAddrs, 
                           const string *namePrefix,
                           Address loadAddress, 
                           bool parseGaps = false)
{
    // Copy the region to a temp file
    void *regionBuf = malloc(regionEnd - regionStart);
    if (!bpProc->lowlevel_process()->readDataSpace
          ((void*)regionStart, regionEnd-regionStart, regionBuf, false)) {
        fprintf(stderr, "%s[%d]: Failed to read from region [%lX %lX]\n",
                __FILE__, __LINE__,(long)regionStart, 
                (long)regionEnd);
    }
    char regName[64];
    if (namePrefix) {
        sprintf(regName, "%s_%lX_%lX", namePrefix->c_str(), regionStart, regionEnd);
    } else {
        sprintf(regName, "MemRegion_%lX_%lX", regionStart, regionEnd);
    }
    FILE *regionFD = fopen(regName, "wb");
    if (regionFD == NULL) {
        fprintf(stderr,"%s[%d]Was unable to open file %s for writing\n",
                FILE__, __LINE__, regName);
        return NULL;
    }
    unsigned writeBytes = fwrite(regionBuf, 1, regionEnd-regionStart, regionFD);
    if (writeBytes != regionEnd-regionStart) {
        fprintf(stderr, "%s[%d] Wrote 0x%x bytes to file %s"
                "when writing region [%lx %lx] to disk\n",
                FILE__,__LINE__,writeBytes, regName, (long)regionStart, (long)regionEnd);
    }
    fclose(regionFD);
    free(regionBuf);

    // create a fileDescriptor and mapped_object and add region as shared object
    fileDescriptor fdesc;
    if (loadAddress == 0) {
        loadAddress = regionStart;
    }
    fdesc = fileDescriptor(regName, regionStart - loadAddress, 
                           regionEnd - regionStart, loadAddress);
    mapped_object *obj = mapped_object::createMappedObject
        (fdesc, bpProc->lowlevel_process(), parseGaps);

    // add function stubs for the entry points we found
    for (unsigned int i=0; i < funcEntryAddrs.size(); i++) {
        obj->parse_img()->addFunctionStub(funcEntryAddrs[i] - loadAddress);
    }

    // force mapped_object to be parsed and return BPatch_module for it
    obj->analyze(); 
    bpProc->lowlevel_process()->addASharedObject(obj);
    pdvector<mapped_module*> mods = obj->getModules();
    if (mods.size() == 0) {
        fprintf(stderr,"%s[%d] Failed to create module for region [%lx %lx]\n",
                FILE__, __LINE__, regionStart, regionEnd);
        return NULL;
    }
    return new BPatch_module(bpProc, mods[0]->proc(), mods[0], bpProc->getImage());
}

/* Object re-parsed, if new modules need to be created in order to do
 * so they are added to the affectedModules vector.  This function should
 * only be invoked if all funcEntryAddrs lie within the boundaries of
 * the module.  A true return value means that an EXISTING object was
 * re-parsed via the addition of new code, a false return value means
 * that this was not the case.  
*/
bool reparseObject(BPatch_addressSpace *addSpace, 
                   BPatch_Vector<BPatch_module*> &affectedModules,
                   mapped_object *obj, vector<Address> &funcEntryAddrs)
{
    // parse all functions with valid entry addresses in the module
    bool reparsedObject = false;
    Address baseAddress = obj->parse_img()->desc().loadAddr();
    vector<Address>::iterator curEntry = funcEntryAddrs.begin();
    Region *reg;
    vector<image_func*> regFuncs;
    while (curEntry != funcEntryAddrs.end()) {
        reg = obj->parse_img()->getObject()->findEnclosingRegion
            ((*curEntry)-baseAddress);
        // if the address is in some region of the binary
        // and isCode is true, meaning that that code was mapped into
        // the binary from disk, analyze the function entry
        if (reg != NULL && obj->parse_img()->isCode((*curEntry) - baseAddress)) {
            image_func *imgfunc = obj->parse_img()->addFunctionStub
                ((*curEntry) - baseAddress);
            regFuncs.clear();
            regFuncs.push_back(imgfunc);
            obj->analyzeNewFunctions(&regFuncs);
            curEntry = funcEntryAddrs.erase(curEntry);
            reparsedObject = true;
        } else {
            curEntry++;
        }
    }

    // return now if there are no more unresolved funcEntryAddrs
    if (funcEntryAddrs.size() == 0) {
        return reparsedObject;
    }
    // get the binary's code regions from symtabAPI or return 
    std::vector<Region*> regions;
    if( ! obj->parse_img()->getObject()->getCodeRegions(regions) ) {
        fprintf(stderr,"%s[%d] Found no code regions for object %s\n", 
                __FILE__,__LINE__, obj->parse_img()->pathname().c_str());
        return false;
    }
    vector<Address> regEntries;
    // for each region, determine which entry addrs lie within it
    for (unsigned ridx=0; funcEntryAddrs.size() && ridx < regions.size(); ridx++) { 
        regEntries.clear();
        // if entry point in region, pass from funcEntryAddrs to regEntries
        curEntry = funcEntryAddrs.begin();
        while (curEntry != funcEntryAddrs.end()) {
            if ((*curEntry) >=(regions[ridx]->getRegionAddr() + baseAddress) &&
                (*curEntry) < (regions[ridx]->getRegionAddr() + 
                               regions[ridx]->getMemSize() + baseAddress)) {
                regEntries.push_back(*curEntry);
                curEntry = funcEntryAddrs.erase(curEntry);
            } else {
                curEntry++;
            }
        }
        if (regEntries.size()) {
            // This case only happens on PE binaries.  We would like
            // to re-parse the region, but we already know that the
            // address lies in an unmapped memory region and SymtabAPI
            // doesn't have memory allocated for those addresses.  If
            // there's code here it didn't come from the binary on
            // disk, so the region was overwritten dynamically and the
            // data will have to be fetched from memory.
            Address regionStart;
            Address regionSize;
            unsigned long rawDataSize = regions[ridx]->getDiskSize();
            if (rawDataSize == 0 || regions[ridx]->getPtrToRawData() == NULL) {
                // Case 1: There is no allocated memory for the region
                // Set rawDataPtr for the region to a newly malloc'ed
                // region and trigger parsing on the region.  
                regionStart = regions[ridx]->getRegionAddr() + baseAddress;
                regionSize = regions[ridx]->getMemSize();
                void* regBuf = malloc (regionSize);
                if (!((BPatch_process*)addSpace)->lowlevel_process()->
                    readDataSpace((void*)regionStart, regionSize, regBuf, true)) {
                    fprintf(stderr, "%s[%d] Failed to read from region [%lX %lX]\n",
                            __FILE__, __LINE__,(long)regions[ridx]->getRegionAddr(), 
                            regions[ridx]->getMemSize());
                    free(regBuf);
                    continue;
                }
                regions[ridx]->setPtrToRawData(regBuf,regions[ridx]->getMemSize());
                curEntry = regEntries.begin();
                regFuncs.clear();
                while (curEntry != regEntries.end()) {
                    image_func *imgfunc = obj->parse_img()->addFunctionStub
                        ((*curEntry) - baseAddress);
                    regFuncs.push_back(imgfunc);
                    curEntry++;
                }
                obj->analyzeNewFunctions(&regFuncs);
                reparsedObject = true;
            } else {
                // Case 2: The points lie past the end of mapped region memory
                // Create a new module corresponding to the gap between
                // the end of the mapped memory for the region and the
                // end of the region boundary, and parse it. 
                // KEVINTODO: a better approach for this case would be to
                // malloc a new region and copy over the mapped part and
                // grab the rest from the mutatee's address space, but I'm
                // not sure what would happen if I did that.
                regionStart = baseAddress + regions[ridx]->getDiskSize();
                regionSize = baseAddress + regions[ridx]->getMemSize()
                    - regionStart;
                BPatch_module * mod = parseRegion((BPatch_process*)addSpace, 
                                                  regionStart,
                                                  regionStart + regionSize,
                                                  regEntries,
                                                  &obj->parse_img()->desc().file(),
                                                  obj->parse_img()->desc().loadAddr());
                affectedModules.push_back(mod);
            }
        }
    }
    return reparsedObject;
}

/* This function is only called when functionEntryAddrs contains
 * entries that do not overlap any existing mapped_object.  The
 * function's implementation depends on the second assumption.
 * Searches /proc for regions that might have mapped in.  The linker
 * tables changes are immediately handled and the mapped object list
 * gets updated, so we don't need to search for them.
 */
#if defined(os_linux) //search /proc for regions
void findNewEnclosingRegions(BPatch_Vector<Address> &funcEntryAddrs, 
                          vector<Address> &regionStartVec, 
                          vector<Address> &regionEndVec, 
                          vector<vector<Address> > &regionEntryPoints,
                          vector<string*> &regionNames, 
                          process *proc)
{
    vector<Address>::iterator curEntry;
    unsigned maps_size=0; 
    map_entries *maps = getLinuxMaps(proc->getPid(), maps_size);
    if (maps) {
        for (unsigned i=0; i < maps_size; i++) {
            curEntry = funcEntryAddrs.begin();
            bool firstEntryInRegion = true;
            while (curEntry != funcEntryAddrs.end()) {
                if (maps[i].start <= (*curEntry) &&
                    maps[i].end > (*curEntry)) {
                    if (firstEntryInRegion) {
                        regionStartVec.push_back(maps[i].start);
                        regionEndVec.push_back(maps[i].end);
                        regionEntryPoints.push_back(vector<Address>());
                        regionNames.push_back(new string(maps[i].path));
                        firstEntryInRegion = false;
                    }
                    regionEntryPoints.back().push_back(*curEntry);
                    curEntry = funcEntryAddrs.erase(curEntry);
                }
                else {
                    curEntry++;
                }
            }
        }
        free(maps);
    }
    return;
}
#else
void findNewEnclosingRegions(BPatch_Vector<Address> &, 
                             vector<Address> &,
                             vector<Address> &,
                             vector<vector<Address> > &,
                             vector<string*> &,
                             process *)
{
}
#endif

/* BPatch_image::parseNewFunctions
 *
 * Uses function entry addresses to trigger the parsing of new
 * functions.  It may re-trigger parsing on existing modules or create
 * new ones to be parsed. 
 *
 * (full description in ../h/BPatch_image.h)
 */
bool BPatch_image::parseNewFunctionsInt
     (BPatch_Vector<BPatch_module*> &affectedModules, 
      const BPatch_Vector<Dyninst::Address> &funcEntryAddrs)
{
    if (!addSpace->getType() == TRADITIONAL_PROCESS) {
        fprintf(stderr,"%s[%d] ERROR: parseNewFunctions has only been "
                "implemented for live processes\n", __FILE__, __LINE__);
        return false;
    }
    if (funcEntryAddrs.size() == 0) {
        fprintf(stderr,"%s[%d] parseNewFunctions requires a non-empty vector "
                "of function entry points\n", __FILE__, __LINE__);
        return false;
    }

    std::vector<AddressSpace*> asv;
    addSpace->getAS(asv);
    AddressSpace *as = asv[0];
    pdvector<mapped_object *> allobjs = as->mappedObjects();

    // group funcEntryAddrs by the mapped_objects that they fit into
    vector<Address> funcEntryAddrs_(funcEntryAddrs); // make an editable copy
    vector<Address>::iterator curEntry;
    vector<Address> objEntries;
    for (unsigned int i=0; i < allobjs.size() && funcEntryAddrs_.size(); i++) { 
        image *curImg = allobjs[i]->parse_img();
        // determine region start/end boundaries
        Address regStart, regEnd;
        Address baseAddress = curImg->desc().loadAddr();
        // iterate through executable regions
        std::vector<Region*> regions;
        if( ! curImg->getObject()->getCodeRegions(regions) ) {
            fprintf(stderr,"%s[%d] Found no code regions for object %s\n", 
                    __FILE__,__LINE__, curImg->pathname().c_str());
            break;
        }
        std::vector<Region *>::const_iterator curreg = regions.begin();
        while(curreg != regions.end() && funcEntryAddrs_.size()) {
            regStart = baseAddress + (*curreg)->getRegionAddr();
            regEnd = regStart + (*curreg)->getMemSize();
            // iterate through func entry addrs
            curEntry = funcEntryAddrs_.begin();
            while (curEntry != funcEntryAddrs_.end()) {
                if ((*curEntry) >= regStart && (*curEntry) < regEnd) {
                    objEntries.push_back(*curEntry);
                    curEntry = funcEntryAddrs_.erase(curEntry);
                } else {
                    curEntry++;
                }
            }
            curreg++;
        }
        // re-parse object if it contains entry points
        if (objEntries.size()) {
            if (reparseObject(addSpace, affectedModules, 
                              allobjs[i], objEntries)) {
                pdvector<mapped_module*> mods = allobjs[i]->getModules();
                //reparseObject adds all new functions to mods[0]
                // since that's what image::addFunctionStub does
                affectedModules.push_back(findOrCreateModule(mods[0]));
            }
            objEntries.clear();
        }
    }
    // if entry points left over, find regions that contain them
    // and parse each region
    if (funcEntryAddrs_.size()) {
        vector<Address> regionStartVec, regionEndVec;
        vector<vector<Address> > regionEntries;
        vector<string*> regionNames;
        findNewEnclosingRegions(funcEntryAddrs_, 
                             regionStartVec, 
                             regionEndVec, 
                             regionEntries, 
                             regionNames,
                             (process*)as);
        for (unsigned ridx=0; ridx < regionStartVec.size(); ridx++) {
            BPatch_module *mod = parseRegion((BPatch_process*)addSpace, 
                                             regionStartVec[ridx], 
                                             regionEndVec[ridx],
                                             regionEntries[ridx],
                                             regionNames[ridx],0);
            if (mod) {
                affectedModules.push_back(mod);
            }
            free(regionNames[ridx]);
            regionNames[ridx] = NULL;
        }
    }
    if (affectedModules.size() == 0) {
        return false;
    }
    return true;
}

// Merges unresolved control flow vectors of all mapped objects return
// the merged vector of all unresolved control flow do this by
// grabbing one module per mapped object, and calling
// BPatch_module::getUnresolvedControlFlow on that module
BPatch_Vector<BPatch_point *> *BPatch_image::getUnresolvedControlFlowInt()
{
   unresolvedCF.clear();
   pdvector<mapped_object *> objs;
   
   std::vector<AddressSpace *> asv;
   addSpace->getAS(asv);
   for (unsigned i=0; i<asv.size(); i++) {
      const pdvector<mapped_object *> &mobjs = asv[i]->mappedObjects();
      for (unsigned j=0; j<mobjs.size(); j++) {
         objs.push_back(mobjs[j]);
      }
   }

   for (unsigned i=0; i < objs.size(); i++) {
      pdvector<mapped_module *> mods = objs[i]->getModules();
      if (mods.size()) {
         BPatch_Vector<BPatch_point*> *badCF = 
             findOrCreateModule(mods[0])->getUnresolvedControlFlow();
         for (unsigned j=0; j < badCF->size(); j++) {
            unresolvedCF.push_back((*badCF)[j]);
         }
      }
   }
   return &unresolvedCF;
}

void BPatch_image::removeAllModules()
{
   BPatch_Vector<BPatch_module *>::iterator i;
   for (i = modlist.begin(); i != modlist.end(); i++)
   {
      (*i)->handleUnload();
   }
   modlist.clear();
}

void BPatch_image::removeModule(BPatch_module *mod) 
{
#if !defined(USE_DEPRECATED_BPATCH_VECTOR)
   modlist.erase(std::find(modlist.begin(),
            modlist.end(),
            mod));
#else
   for (unsigned j = 0; j < modlist.size(); j++) {
      if (modlist[j] == mod) {
         modlist.erase(j);
      }
   }
#endif
   mod->handleUnload();
}

bool BPatch_image::readStringInt(BPatch_variableExpr *expr, std::string &str,
                                 unsigned size_limit)
{
   const BPatch_type *type = expr->getType();
   if (!type) {
      bperr("String read attempted on variable with no type information");
      return false;
   }
   
   Address addr = 0x0;
   if (type->getDataClass() == BPatch_dataPointer) {
      void *value;
      expr->readValue(&value);
      addr = (Address) value;
   }
   else if (type->getDataClass() == BPatch_dataArray) {
      addr = (Address) expr->getBaseAddr();
   }
   else {
      bperr("String read failed on variable with unexpected type");
      return false;
   }
   if (!addr)
      return false;

   return readString(addr, str, size_limit);
}


bool BPatch_image::readStringInt(Address addr, std::string &str, unsigned size_limit)
{
   char *buffer = NULL;
   unsigned buffer_offset = 0;
   unsigned buffer_size = 0;
   bool result;
   bool should_continue = false;

   BPatch_process *proc = dynamic_cast<BPatch_process *>(getAddressSpace());
   if (!proc) {
      //Only works with dynamic processes;
      return false;
   }
   if (proc && !proc->isStoppedInt()) {
      should_continue = true;
      proc->stopExecution();
   }

   std::vector<AddressSpace *> asv;
   getAddressSpace()->getAS(asv);
   AddressSpace *as = asv[0];
   assert(as);

   unsigned word_size = as->getAddressWidth();
   assert(word_size == 4 || word_size == 8);
   
   Address start_word = (word_size == 4) ? (addr & ~0x3) : (addr & ~0x7);
   unsigned start_offset = addr - start_word;

   for (;;) {
      if (!buffer) {
         buffer_size = 256;
         buffer = (char *) malloc(buffer_size);
         assert(buffer);
      }
      if (buffer_offset + word_size + 1 > buffer_size) {
         buffer_size *= 2;
         buffer = (char *) realloc(buffer, buffer_size);
         assert(buffer);
      }

      result = as->readDataSpace((void *) (start_word + buffer_offset), word_size, 
                                 buffer + buffer_offset, false);
      if (!result) {
         signal_printf("[%s:%u] - ERROR reading address %x for string\n",
                       FILE__, __LINE__, start_word + buffer_offset);
         bperr("Error reading from target process");
         goto done;
      }

      buffer_offset += word_size;

      if (size_limit && 
          size_limit < buffer_offset - start_offset) {
         buffer[size_limit + start_offset] = '\0';
         signal_printf("[%s:%u] - WARN string read at %x exceeded size limit of %d",
                       FILE__, __LINE__, addr, size_limit);
         bpwarn("String read exceeded size limit");
         break;
      }

      bool done = false;
      for (unsigned i=1; i<=word_size; i++) {
         if (buffer_offset-i < addr-start_word)
            continue;
         if (buffer[buffer_offset-i] == '\0') {
            done = true;
            break;
         }
      }
      if (done)
         break;
   }

   str = buffer + start_offset;
   result = true;

 done:
   if (buffer)
      free(buffer);
   if (should_continue) {
      assert(proc);
      proc->continueExecution();
   }
   
   return result;
}

